/*************************************************************************/
/*************************************************************************/
/*                                                                       */
/*  File Name : AirScheduleSolu2.h                                       */
/*                                                                       */
/*       Date : November  3, 1997                                        */
/* Updated at : August   22, 1998                                        */
/*                                                                       */
/*     Author : H. Timucin OZDEMIR                                       */
/*                                                                       */
/*************************************************************************/
/*************************************************************************/

#ifndef _AIRSCHEDULESOLU2_H
#define _AIRSCHEDULESOLU2_H

#include <iostream.h>
#include <math.h>

#include <tSet.h>
#include <gVector.h>

#include "AirScheduleProblem.h"
#include "ListOfEdgeLists.h"
#include "Bin.h"
#include "Pair.h"
#include "CandidateELSet.h"
#include "Cover.h"
#include "debugs.h"

class AirScheduleSolu2
{  public :

     AirScheduleSolu2(AirScheduleProblem*);
     AirScheduleSolu2(AirScheduleProblem*,EdgeGraph*);
     AirScheduleSolu2(const AirScheduleSolu2&);
     AirScheduleSolu2();
     ~AirScheduleSolu2();

     AirScheduleProblem* getProblem() const;
     EdgeGraph* getEdgeGraph() const;
     double getCost() ;
     double getCost2() ;
     double getDCost() ;
     double getRotationSTD(double&,double&,double&,int&,int&,int&) ;

     // Initializer
     void solveRandom();
     void solveRandom2();
     void solveGreedy();
     void solveGreedy2();
     void solveGreedyBin();

     // Baldwin Effect - A Simple LocalSearch
     void applyLocalSearch();
     void applyLocalSearch2();
     void applyLocalSearch3();

     void setProblem(AirScheduleProblem*);

     void discardListOfEdgeLists(int);
     void discardAllListOfEdgeLists();

     inline getNumOfRotations() const 
     { return myEdgeLists.getIndex(); } ;

     inline getNumberOfFlights() const 
     { return (myProblem!=NULL?myProblem->getNumberOfFlights():1) ; } ;


     getNumberOfBaseCities() ;
     getNumberOfUncoveredNodes() ;


     inline double getSquareSumOfRotationFlyingTimes() 
	    { return getSquareSumOfRotationFlyingTimes(100.0); } ;
     double getSquareSumOfRotationFlyingTimes(float param) ;
     double getUnderUtilizationRatio() ; 

     // Returns the given ListOfEdheLists object
     inline ListOfEdgeLists *getListOfEdgeListsIndex(int indx)
     { return myEdgeLists.getNthElement(indx); } ;

     // Inserts a ListOfEdgeLists
     void insertEdgeList(EdgeList);
     void addListOfEdgeList(ListOfEdgeLists);
     void patchUp(gVector<Edge*>* abondonedEdgeSet=NULL);

     canMerge(EdgeList);

     void writeStat(ostream&);

     friend ostream & operator<<(ostream&,AirScheduleSolu2&);
     operator==(const AirScheduleSolu2&);
     AirScheduleSolu2 operator=(const AirScheduleSolu2&);

     findCover(int NumOfFlights, char * covered) ;
     void produceColumns();

     EdgeList* decideEdgeList(EdgeList* currELst, EdgeGraph *currEG, 
			      EdgeGraph *origEG, char * covered, int
			      DEBUG=0);

     findAllPaths(EdgeList* currELst, Pair & pair, int DEBUG=0);
     findMaximumPath(EdgeList* currELst, Pair & pair, int STRICT, int DEBUG=0);

     putEdgeList(EdgeList *crrEdgeList, EdgeGraph *currEG, char *covered);

     void getScheduleFromColumnsFile(char*);

     void writeCoded(ostream& os);
     void readCoded(istream& is);

     EdgeList getAllEdges();

   protected:

     gVector<ListOfEdgeLists>  myEdgeLists;
                 double  cost;
     AirScheduleProblem  *myProblem;
		   char  calculated;
     EdgeGraph           *edgeGraph;
		   char  debug; 
     CandidateELSet      *candidates;
                 double  minRotationTime;
                 double  maxRotationTime;
                 double  aveRotationTime;
                 double  stdRotationTime;
                    int  notEndsWStarted;
                    int  notFromBC;
                    int  numOfDeadheadings;
                   char  calculatedStats;
                 double  myUURatio;
                   char  calculatedUUR;
                  Cover  *myCover;
                 double  tofSpanningTimes;
                 double  tofFlyingTimes;
                   char  calculatedCover;

     
     void sortDes();
     void bubbleUpColumnsAfter(unsigned indx);

     void dispose();
     void copyFrom(const AirScheduleSolu2&);

     void removeEdgesStartOrEndFrom(int NumOfFlights,
		       WeightedRandomPickerSet<Edge*> *Es, char * covered);

     addEdgesFromSet(int NumOfFlights, EdgeGraph *currEG, EdgeGraph *origEG, 
		     WeightedRandomPickerSet<Edge*> *Es, char * covered,
		     int DEBUG=0);

     addEdgesFromSetR(int NumOfFlights, EdgeGraph *currEG, EdgeGraph *origEG, 
		     WeightedRandomPickerSet<Edge*> *Es, char * covered,
		     int DEBUG=0);

     addEdgesFromSetR_U(int NumOfFlights, EdgeGraph *currEG, EdgeGraph *origEG, 
		     WeightedRandomPickerSet<Edge*> *Es, char * covered,
		     int UpdateSolutions, int DEBUG=0);

     void addEdgesForPatchUp( int NumOfFlights, EdgeGraph *currEG, 
			      char * covered, float preUsageProbability=0.3);

     Edge** pickEdgeFromCovers(WeightedRandomPickerSet<Edge*> *Es,
					  tSet<int> *res);

     Edge** semiRandomPick(int NumOfFlights, EdgeGraph *currEG,
		  WeightedRandomPickerSet<Edge*> *Es, char * covered);


     /** Returns the address of the Edge address if there is an edge */
     /** it starts from/ends at res                                  */
     EdgeList pickEdgeFromSetHaving(WeightedRandomPickerSet<Edge*> *Es, int res);

     void addEdgeToListOfEdgeList(Edge* edge, char* covered);

     WeightedRandomPickerSet<int> getUncoveredNodeSet(int NumOfFlights,
                                                      char* covered);

     void removeNodeFromListOfEdgeLists(int NumberOfFlights,int otherNode,
		 EdgeGraph * origEG, EdgeGraph * currEG, char *covered);

     //Edge** 
     Edge* 
     isThereUncoveredMatch(int NumOfFlights, 
	     WeightedRandomPickerSet<Edge*> *potential, char* covered,
	     int DEBUG=0);

     EdgeList*
     isThereUncoveredMatchPath(int NumOfFlights, EdgeGraph *origEG,
	      WeightedRandomPickerSet<Edge*> *potential, char* covered,
	      int DEBUG=0);

     void buildCandidates( int NumOfFlights, EdgeGraph *origEG,
	                   char* covered, int STRICT, int DEBUG );

     pickPath( int NumOfFlights, EdgeGraph *origEG, char* covered,
	       EdgeList& result, int STRICT, int DEBUG=0 );

     HeuristicPatchUpAtMost(int NumOfFlights, EdgeGraph *origEG, 
			   EdgeGraph *currEG, char * covered, 
			   int numberOfTrials, int DEBUG=0);

     HeuristicPatchUp( int NumOfFlights, EdgeGraph *origEG, 
	               EdgeGraph *currEG, char * covered, int DEBUG=0)
     { return HeuristicPatchUpAtMost(NumOfFlights,origEG,currEG,covered,NUM_OF_HEU_TRIALS,DEBUG);};

     //void findCover(int NumOfFlights, char * covered);
     void printCover(int NumOfFlights, char * covered) 
	  { printCover(cout,NumOfFlights,covered); } ;
     void printCover(ostream& os,int NumOfFlights, char * covered) const;

     void nicePrint(ostream& os) ;

     void binary(char*,int,int,int);

     WeightedRandomPickerSet<Edge*>* filterEdgesNotStartFromBase(WeightedRandomPickerSet<Edge*> *Es);

     void putEdgesBackFromEdgeList(EdgeList *crrEdgeList, EdgeGraph *currEG, 
                                   EdgeGraph *origEG, char *covered);

     solveRandomStep1();

     void applyRotationCheck(int NumOfFlights, Bin *myBins, int binSize);

     void checkMem(char*);

     void removeRedundancy();

     void patchUpWithDeadHeadings(int NumOfFlights, char *cover);
     void patchUpWithDeadHeadings1(int NumOfFlights, char *cover);

     void myReset();
     void CalculateStatistics();
     void updateProbabilities(WeightedRandomPickerSet<Edge*> *Es);
     doesCoverUncovered(Cover* cvr, EdgeList* edgeList);


}; // end of class AirScheduleSolu2

#endif
